/*
 * IR toolkit for tcl/tk
 * (c) Index Data 1995
 * See the file LICENSE for details.
 * Sebastian Hammer, Adam Dickmeiss
 *
 * tkinit.c,v
 * Revision 1.12  1996/07/03  13:31:14  adam
 * The xmalloc/xfree functions from YAZ are used to manage memory.
 *
 * Revision 1.11  1996/06/27  14:21:02  adam
 * Yet another Windows port.
 *
 * Revision 1.10  1996/03/27  17:00:55  adam
 * Fix: main defined when using Tk3.6; it shouldn't be.
 *
 * Revision 1.9  1996/03/20  13:54:05  adam
 * The Tcl_File structure is only manipulated in the Tk-event interface
 * in tkinit.c.
 *
 * Revision 1.8  1996/02/21  10:16:22  adam
 * Simplified select handling. Only one function ir_tcl_select_set has
 * to be externally defined.
 *
 * Revision 1.7  1996/02/05  17:58:05  adam
 * Ported ir-tcl to use the beta releases of tcl7.5/tk4.1.
 *
 * Revision 1.6  1996/01/04  11:05:24  adam
 * New setting: PDUType - returns type of last PDU returned from the target.
 * Fixed a bug in configure/Makefile.
 *
 * Revision 1.5  1995/11/28  13:52:33  quinn
 * Windows port.
 *
 * Revision 1.4  1995/09/21  13:42:55  adam
 * Bug fixes.
 *
 * Revision 1.3  1995/09/21  13:11:53  adam
 * Support of dynamic loading.
 * Test script uses load command if necessary.
 *
 * Revision 1.2  1995/06/29  12:34:09  adam
 * IrTcl now works with both tk4.0b4/tcl7.4b4 and tk3.6/tcl7.3
 *
 * Revision 1.1  1995/06/26  10:20:20  adam
 * ir-tk works like wish.
 *
 *
 * Revision 1.17  1995/06/21  11:04:55  adam
 * Uses GNU autoconf 2.3.
 * Install procedure implemented.
 * boook bitmaps moved to sub directory bitmaps.
 *
 * Revision 1.16  1995/06/01  16:36:47  adam
 * About buttons. Minor bug fixes.
 *
 * Revision 1.15  1995/05/31  13:10:00  adam
 * Client searches/presents may be interrupted.
 * New moving book-logo.
 *
 * Revision 1.14  1995/05/29  08:44:30  adam
 * Work on delete of objects.
 *
 * Revision 1.13  1995/05/26  11:44:11  adam
 * Bugs fixed. More work on MARC utilities and queries. Test
 * client is up-to-date again.
 *
 * Revision 1.12  1995/05/23  15:34:49  adam
 * Many new settings, userInformationField, smallSetUpperBound, etc.
 * A number of settings are inherited when ir-set is executed.
 * This version is incompatible with the graphical test client (client.tcl).
 *
 * Revision 1.11  1995/05/11  15:34:47  adam
 * Scan request changed a bit. This version works with RLG.
 *
 * Revision 1.10  1995/03/21  13:41:04  adam
 * Comstack cs_create not used too often. Non-blocking connect.
 *
 * Revision 1.9  1995/03/20  08:53:31  adam
 * Event loop in tclmain.c rewritten. New method searchStatus.
 *
 * Revision 1.8  1995/03/17  18:26:18  adam
 * Non-blocking i/o used now. Database names popup as cascade items.
 *
 * Revision 1.7  1995/03/17  07:50:33  adam
 * Headers have changed a little.
 */

#include <tk.h>
#include <log.h>
#include "ir-tcl.h"

#if TK_MAJOR_VERSION > 4 || (TK_MAJOR_VERSION == 4 && TK_MINOR_VERSION > 0)
#define IRTCL_GENERIC_FILES 1
#else
#define IRTCL_GENERIC_FILES 0
#endif

#if WINDOWS
#define IRTCL_USE_TIMER 1
#else
#define IRTCL_USE_TIMER 0
#endif

struct sel_proc {
    void (*f)(ClientData clientData, int r, int w, int e);
    ClientData clientData;
    int fd;
#if IRTCL_USE_TIMER
    int mask;
    Tcl_TimerToken timer_token;
#else
#if IRTCL_GENERIC_FILES
    Tcl_File tcl_File;
#endif
#endif
    struct sel_proc *next;
};

static struct sel_proc *sel_proc_list = NULL;

#if IRTCL_USE_TIMER
static void ir_tcl_timer_proc (ClientData clientData)
{
    struct sel_proc *sp = (struct sel_proc *) clientData;

    if (!sp->f)
        return ;
    sp->timer_token =
        Tcl_CreateTimerHandler (250, ir_tcl_timer_proc, clientData);
    (*sp->f)(sp->clientData, sp->mask & TK_READABLE, sp->mask & TK_WRITABLE,
             sp->mask & TK_EXCEPTION);
    
}

void ir_tcl_select_set (void (*f)(ClientData clientData, int r, int w, int e),
                        int fd, ClientData clientData, int r, int w, int e)
{
    int mask = 0;
    struct sel_proc **sp = &sel_proc_list;

    if (r)
        mask |= TK_READABLE;
    if (w)
        mask |= TK_WRITABLE;
    if (e)
        mask |= TK_EXCEPTION;
    while (*sp)
    {
        if ((*sp)->fd == fd)
            break;
        sp = &(*sp)->next;
    }
    if (!*sp)
    {
        *sp = ir_tcl_malloc (sizeof(**sp));
        (*sp)->next = NULL;
        (*sp)->fd = fd;
        (*sp)->timer_token =
            Tcl_CreateTimerHandler (250, ir_tcl_timer_proc, *sp);
    }
    (*sp)->mask = TCL_READABLE|TCL_WRITABLE;
    (*sp)->f = f;
    (*sp)->clientData = clientData;
    if (!f)
    {
        struct sel_proc *sp_tmp = *sp;
        Tcl_DeleteTimerHandler ((*sp)->timer_token);
        *sp = (*sp)->next;
        xfree (sp_tmp);
    }
}

#else
static void ir_tcl_tk_select_proc (ClientData clientData, int mask)
{
    struct sel_proc *sp = (struct sel_proc *) clientData;

    if (!sp->f)
        return ;
    (*sp->f)(sp->clientData, mask & TK_READABLE, mask & TK_WRITABLE,
             mask & TK_EXCEPTION);
}

void ir_tcl_select_set (void (*f)(ClientData clientData, int r, int w, int e),
                        int fd, ClientData clientData, int r, int w, int e)
{
    int mask = 0;
    struct sel_proc *sp = sel_proc_list;

    if (r)
        mask |= TK_READABLE;
    if (w)
        mask |= TK_WRITABLE;
    if (e)
        mask |= TK_EXCEPTION;
    while (sp)
    {
        if (sp->fd == fd)
             break;
        sp = sp->next;
    }
    if (!sp)
    {
        sp = ir_tcl_malloc (sizeof(*sp));
        sp->next = sel_proc_list;
        sel_proc_list = sp;
        sp->fd = fd;
#if IRTCL_GENERIC_FILES
#if WINDOWS
        sp->tcl_File = Tcl_GetFile (fd, TCL_WIN_SOCKET);
#else
        sp->tcl_File = Tcl_GetFile ((ClientData) fd, TCL_UNIX_FD);
#endif
#endif
    }
    sp->f = f;
    sp->clientData = clientData;
#if IRTCL_GENERIC_FILES
    if (f)
        Tcl_CreateFileHandler (sp->tcl_File, mask, ir_tcl_tk_select_proc, sp);
    else
        Tcl_DeleteFileHandler (sp->tcl_File);
#else
    if (f)
        Tk_CreateFileHandler (fd, mask, ir_tcl_tk_select_proc, sp);
    else
        Tk_DeleteFileHandler (fd);
#endif
}
#endif

#if WINDOWS || TK_MAJOR_VERSION == 3

#else
int main (int argc, char **argv)
{
    Tk_Main (argc, argv, Tcl_AppInit);
    return 0;
}
#endif

int Tcl_AppInit (Tcl_Interp *interp)
{
#if TK_MAJOR_VERSION != 4 || TK_MINOR_VERSION < 1
    Tk_Window main;

    main = Tk_MainWindow(interp);
    if (main == NULL)
        return TCL_ERROR;
#endif
    if (Tcl_Init(interp) == TCL_ERROR)
        return TCL_ERROR;
    if (Tk_Init(interp) == TCL_ERROR)
        return TCL_ERROR;
#if TK_MAJOR_VERSION ==4 && TK_MINOR_VERSION >= 1
    Tcl_StaticPackage (interp, "Tk", Tk_Init, (Tcl_PackageInitProc *) NULL);
#endif
    if (Irtcl_Init(interp) == TCL_ERROR)
        return TCL_ERROR;
    return TCL_OK;
}
